#include <stdbool.h>
#include <sys/param.h> /* for MIN/MAX */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <limits.h>

#include "exanic.h"
#include "pcie_if.h"
#include "port.h"
#include "config.h"
#include "util.h"

int exanic_check_supported(exanic_t *exanic)
{
    uint32_t pcie_if_ver;
    exanic_hardware_id_t hw_id;
    exanic_function_id_t func_id;

    pcie_if_ver = exanic->registers[REG_EXANIC_INDEX(REG_EXANIC_PCIE_IF_VER)];
    hw_id = exanic->registers[REG_EXANIC_INDEX(REG_EXANIC_HW_ID)];
    func_id = exanic->registers[REG_EXANIC_INDEX(REG_EXANIC_FUNCTION_ID)];

    if (pcie_if_ver != 1)
    {
        exanic_err_printf("unknown exanic interface version: %u", pcie_if_ver);
        return -1;
    }

    if (exanic_hardware_id_str(hw_id) == NULL)
    {
        exanic_err_printf("unknown hardware ID: %u", hw_id);
        return -1;
    }

    if (exanic_function_id_str(func_id) == NULL)
    {
        exanic_err_printf("unknown exanic function ID: %u", func_id);
        return -1;
    }

    return 0;
}

exanic_hardware_id_t exanic_get_hw_type(exanic_t *exanic)
{
    return exanic->registers[REG_EXANIC_INDEX(REG_EXANIC_HW_ID)];
}

exanic_function_id_t exanic_get_function_id(exanic_t *exanic)
{
    return exanic->registers[REG_EXANIC_INDEX(REG_EXANIC_FUNCTION_ID)];
}

uint32_t exanic_get_caps(exanic_t *exanic)
{
    return exanic->registers[REG_EXANIC_INDEX(REG_EXANIC_CAPS)];
}

time_t exanic_get_hw_rev_date(exanic_t *exanic)
{
    return exanic->registers[REG_EXANIC_INDEX(REG_EXANIC_HW_REV_DATE)];
}

volatile uint32_t *exanic_get_devkit_registers(exanic_t *exanic)
{
    return exanic->devkit_regs_region;
}

char *exanic_get_devkit_memory(exanic_t *exanic)
{
    return exanic->devkit_mem_region;
}

volatile uint32_t *exanic_get_extended_devkit_registers(exanic_t *exanic)
{
    return exanic->devkit_regs_ex_region;
}

char *exanic_get_extended_devkit_memory(exanic_t *exanic)
{
    return exanic->devkit_mem_ex_region;
}

int exanic_is_devkit_demo(exanic_t *exanic)
{
    return exanic->registers[REG_EXANIC_INDEX(REG_EXANIC_DEVKIT_LICENSE_TYPE)] == 1;
}

int exanic_is_devkit_free(exanic_t *exanic)
{
    return exanic->registers[REG_EXANIC_INDEX(REG_EXANIC_DEVKIT_LICENSE_TYPE)] == 2;
}

int exanic_get_sysfs_path(exanic_t *exanic, char *buf, size_t len)
{
    char tmp[PATH_MAX];
    char abs[PATH_MAX];
    char ifname[64];

    if (exanic_get_interface_name(exanic, 0, ifname, sizeof ifname) == -1)
        return -1;

    snprintf(tmp, sizeof tmp, "/sys/class/net/%s/device", ifname);
    if (realpath(tmp, abs) == NULL)
    {
        exanic_err_printf("realpath failed: %s", strerror(errno));
        return -1;
    }

    size_t req = strlen(abs) + 1;
    if (len < req)
    {
        exanic_err_printf("insufficient buffer size: %zu(%zu)",
                          len, req);
        return -1;
    }

    memcpy(buf, abs, req);
    return req - 1;
}

/* generic Common Flash Interface opcodes */
#define CFI_READ_ARRAY               0xFF
#define CFI_QUERY_ADDRESS            0x55
#define CFI_QUERY_DATA               0x98

/* Intel/P30 specific opcodes */
#define P30_SET_CR_SETUP             0x60
#define P30_SET_CR_CONFIRM           0x03
#define P30_CLEAR_STATUS_REG         0x50
#define P30_UNLOCK_BLOCK_SETUP       0x60
#define P30_UNLOCK_BLOCK_CONFIRM     0xD0
#define P30_BLOCK_ERASE_SETUP        0x20
#define P30_BLOCK_ERASE_CONFIRM      0xD0
#define P30_BUFFER_PROGRAM_SETUP     0xE8
#define P30_BUFFER_PROGRAM_CONFIRM   0xD0
/* Intel/P30 status register masks */
#define P30_STATUS_READY_MASK        0x80
#define P30_STATUS_ERROR_MASK        0x30
/* Value to program to P30 config register */
#define P30_CR_CONFIG                0x9803

/**
 * Low-level ExaNIC flash access functions
 */

static void exanic_cfi_flash_init_interface(exanic_t *exanic)
{
    /* nWE high, nCE high, nOE high, nL high */
    exanic->registers[REG_HW_INDEX(REG_HW_FLASH_CTRL)] = 0xf;
    /* dummy read for timing */
    exanic->registers[REG_HW_INDEX(REG_HW_FLASH_CTRL)];
    /* nWE high, nCE low, nOE high, nL low */
    exanic->registers[REG_HW_INDEX(REG_HW_FLASH_CTRL)] = 0x5;
}

static void exanic_cfi_flash_release_interface(exanic_t *exanic)
{
    /* nWE high, nCE high, nOE high, nL high */
    exanic->registers[REG_HW_INDEX(REG_HW_FLASH_CTRL)] = 0xf;
}

static void exanic_cfi_flash_set_address(exanic_t *exanic, uint32_t address)
{
    exanic->registers[REG_HW_INDEX(REG_HW_FLASH_ADDR)] = address;
}

static uint16_t exanic_cfi_flash_read_current(exanic_t *exanic)
{
    uint16_t data;
    /* nWE high, nCE low, nOE high, nL low, flash to FPGA */
    exanic->registers[REG_HW_INDEX(REG_HW_FLASH_CTRL)] = 0x5;
    /* drive OE low */
    exanic->registers[REG_HW_INDEX(REG_HW_FLASH_CTRL)] = 0x1;
    /* dummy read for timing */
    exanic->registers[REG_HW_INDEX(REG_HW_FLASH_DIN_CFI)];
    data = exanic->registers[REG_HW_INDEX(REG_HW_FLASH_DIN_CFI)];
    /* return OE high */
    exanic->registers[REG_HW_INDEX(REG_HW_FLASH_CTRL)] = 0x5;
    return data;
}

static void exanic_cfi_flash_write_current(exanic_t *exanic, uint16_t data)
{
    /* nWE high, nCE low, nOE high, nL low, FPGA to flash */
    exanic->registers[REG_HW_INDEX(REG_HW_FLASH_CTRL)] = 0x15;
    exanic->registers[REG_HW_INDEX(REG_HW_FLASH_DOUT_CFI)] = data;
    /* drive WE low */
    exanic->registers[REG_HW_INDEX(REG_HW_FLASH_CTRL)] = 0x14;
    /* dummy read for timing */
    exanic->registers[REG_HW_INDEX(REG_HW_FLASH_DIN_CFI)];
    /* return WE high */
    exanic->registers[REG_HW_INDEX(REG_HW_FLASH_CTRL)] = 0x15;
}

static uint16_t exanic_cfi_flash_read(exanic_t *exanic, uint32_t address)
{
    exanic_cfi_flash_set_address(exanic, address);
    return exanic_cfi_flash_read_current(exanic);
}

static void exanic_cfi_flash_write(exanic_t *exanic, uint32_t address, uint16_t data)
{
    exanic_cfi_flash_set_address(exanic, address);
    exanic_cfi_flash_write_current(exanic, data);
}

/**
 * P30 flash family specific functions
 */

static void p30_set_asynchronous_mode(exanic_t *exanic)
{
    exanic_cfi_flash_write(exanic, P30_CR_CONFIG, P30_SET_CR_SETUP);
    exanic_cfi_flash_write(exanic, P30_CR_CONFIG, P30_SET_CR_CONFIRM);
}

static bool p30_init(exanic_t *exanic)
{
    exanic_cfi_flash_write(exanic, 0, P30_CLEAR_STATUS_REG);
    return true;
}

static void p30_release(exanic_t *exanic)
{
    exanic_cfi_flash_release_interface(exanic);
}

static bool exanic_cfi_flash_read_multiple(exanic_t *exanic, uint32_t address, uint16_t *data, uint32_t size)
{
    uint32_t curr_addr = address;
    uint32_t i = 0;
    uint16_t *ptr = data;
    
    exanic_cfi_flash_write(exanic, 0, CFI_READ_ARRAY);
    for (; i < size; i++)
        *ptr++ = exanic_cfi_flash_read(exanic, curr_addr++);

    return true;
}

/**
 * Generic flash access functions
 */

int exanic_flash_open_cfi(exanic_t *exanic)
{
    uint16_t command_set;

    exanic_cfi_flash_init_interface(exanic);
    exanic_cfi_flash_write(exanic, CFI_QUERY_ADDRESS, CFI_QUERY_DATA);
    if (exanic_cfi_flash_read(exanic, 0x10) != 'Q')
    {
        /* Flash may be in synchronous read mode */
        p30_set_asynchronous_mode(exanic);
        exanic_cfi_flash_init_interface(exanic);
        exanic_cfi_flash_write(exanic, CFI_QUERY_ADDRESS, CFI_QUERY_DATA);
    }
    if (exanic_cfi_flash_read(exanic, 0x10) != 'Q' ||
        exanic_cfi_flash_read(exanic, 0x11) != 'R' ||
        exanic_cfi_flash_read(exanic, 0x12) != 'Y')
    {
        fprintf(stderr, "ERROR: failed to query flash\n");
        goto error;
    }

    command_set = (exanic_cfi_flash_read(exanic, 0x14) << 8) | (exanic_cfi_flash_read(exanic, 0x13));
    switch (command_set)
    {
        case 1: /* Intel */
        case 2: /* AMD */
            break;
        default:
            fprintf(stderr, "ERROR: unknown flash command set 0x%x\n", command_set);
            goto error;
    }

    return 0;

error:
    exanic_cfi_flash_release_interface(exanic);
    return -1;
}

int exanic_customer_firmware_check(exanic_t *exanic, char *customer, size_t customer_len, char *date, size_t date_len)
{
    uint16_t buffer[256] __attribute__ ((aligned));
    char str_id[16] = {0};
    char time[64] = {0};
    uint32_t address = 0x800000;
    int ret;
    int i;
    
    exanic_flash_open_cfi(exanic);
    p30_init(exanic);
    exanic_cfi_flash_read_multiple(exanic, address, buffer, 256);
    p30_release(exanic);

    for (i=0; i<5; i++) {
        str_id[i*2] = (char)buffer[i];
        str_id[i*2+1] = (char)(buffer[i] >> 8);
    }
    str_id[9] = '\0';

    for (i=5; i<23; i++) {
        time[(i-5)*2] = (char)buffer[i];
        time[(i-5)*2+1] = (char)(buffer[i] >> 8);
    }
    time[35] = '\0';

    if (strcmp(str_id, "AcceleCom") == 0)
        ret = 1;
    else
        ret = 0;

    if (customer)
        strncpy(customer, str_id, MIN(strlen(str_id), customer_len));

    if (date)
        strncpy(date, time, MIN(strlen(time), date_len));
    
    return ret;
}
